@@ -16,14 +16,14 @@ from regexapp.collection import PatternBuilder
 from regexapp.collection import MultilinePattern
 from regexapp.collection import PatternReference
 from regexapp.core import RegexBuilder
-from regexapp.core import DynamicGenTestScript
+from regexapp.core import DynamicTestScriptBuilder
 from regexapp.core import add_reference
 from regexapp.core import remove_reference
 
-__version__ = '0.2.2'
-version = __version__
-__edition__ = 'Community'
-edition = __edition__
+from regexapp.config import version
+from regexapp.config import edition
+__version__ = version
+__edition__ = edition
 
 __all__ = [
     'TextPattern',
@@ -33,7 +33,7 @@ __all__ = [
     'PatternBuilder',
     'PatternReference',
     'RegexBuilder',
-    'DynamicGenTestScript',
+    'DynamicTestScriptBuilder',
     'add_reference',
     'remove_reference',
     'version',
